/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package MatrixSumSolver.fastsumsolver;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 *
 * @author Michael
 */
public class FastSumSolver {

    public void solve(int[][] matrix, int targetSum) {
        long startTime, endTime;

        List<Map<Integer, EquivalenceClass>> equivalenceClasses = new ArrayList<>();

        System.out.println("----------------------------------------------------------------------------");
        System.out.println("fast sum solver, matrix size: "+matrix.length+"x"+matrix[0].length+", target sum = "+targetSum);
        startTime = System.currentTimeMillis();

        // create one map of equivalence classes for each line of the matrix
        for(int l=0; l<matrix.length; l++) {
            Map<Integer, EquivalenceClass> map = new HashMap<>();

            for(int c=0; c<matrix[l].length; c++) {
                int value = matrix[l][c];

                // ignore values > targetSum
                if( value <= targetSum ) {
                    // get equivalence class for this value
                    EquivalenceClass equivalenceClass = map.get(value);
                    if( equivalenceClass == null ) {
                        equivalenceClass = new SingleMatrixElementEquivalenceClass(value);
                        map.put(value, equivalenceClass);
                    }

                    // add value to equivalence class
                    ((SingleMatrixElementEquivalenceClass)equivalenceClass).add(l, c);
                }
            }

            // add map to list of maps
            equivalenceClasses.add(map);
        }

        // now reduce the list of maps by combining classes to new classes
        while( equivalenceClasses.size() > 1 ) {
            List<Map<Integer, EquivalenceClass>> newEquivalenceClasses = new ArrayList<>();

            for(int i=0; i<equivalenceClasses.size(); i+=2) {

                if( i+1 < equivalenceClasses.size() ) {
                    Map<Integer, EquivalenceClass> newMap = new HashMap<>();
                    Map<Integer, EquivalenceClass> map1 = equivalenceClasses.get(i);
                    Map<Integer, EquivalenceClass> map2 = equivalenceClasses.get(i+1);
                    for(int v1 : map1.keySet()) {
                        for(int v2 : map2.keySet()) {
                            int value = v1+v2;
                            if( value <= targetSum ) {
                                EquivalenceClass equivalenceClass = newMap.get(value);
                                if(equivalenceClass == null) {
                                    equivalenceClass = new CombinedEquivalenceClass(value);
                                    newMap.put(value, equivalenceClass);
                                }
                                ((CombinedEquivalenceClass)equivalenceClass).add(map1.get(v1), map2.get(v2));
                            }
                        }
                    }
                    newEquivalenceClasses.add(newMap);
                } else {
                    newEquivalenceClasses.add( equivalenceClasses.get(i));
                }
            }

            equivalenceClasses = newEquivalenceClasses;
        }

        // at last: get equivalence class for the target sum (may be null)
        EquivalenceClass ec = equivalenceClasses.get(0).get(targetSum);

        endTime = System.currentTimeMillis();

        if( ec != null ) {
            System.out.print(ec.getElementCount()+" solutions found");
        } else {
            System.out.print("no solutions found");
        }
        System.out.println( " in "+((endTime-startTime)/1000d)+" seconds");
    }
}
